/***************************************************************************
                          kdiff3.cpp  -  description
                             -------------------
    begin                : Don Jul 11 12:31:29 CEST 2002
    copyright            : (C) 2002 by Joachim Eibl
    email                : joachim.eibl@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
 * $Log$                                                                   *
 ***************************************************************************/

#include "diff.h"

#include <iostream>

// include files for QT
#include <qdir.h>
#include <qprinter.h>
#include <qpainter.h>
#include <qsplitter.h>

// include files for KDE
#include <kiconloader.h>
#include <kmessagebox.h>
#include <kfiledialog.h>
#include <kmenubar.h>
#include <kstatusbar.h>
#include <klocale.h>
#include <kconfig.h>
#include <kstdaction.h>
#include <kcmdlineargs.h>

// application specific includes
#include "kdiff3.h"
#include "optiondialog.h"

#define ID_STATUS_MSG 1

KDiff3App::KDiff3App(QWidget* , const char* name):KMainWindow(0, name)
{
   setCaption( "KDiff3" );
   config=kapp->config();

   m_pBuf1 = 0;
   m_pBuf2 = 0;
   m_pBuf3 = 0;

   m_pMainSplitter = 0;
   m_pDirectoryMergeWindow = 0;
   m_pCornerWidget = 0;
   m_pMainWidget = 0;
   m_pDiffTextWindow1 = 0;
   m_pDiffTextWindow2 = 0;
   m_pDiffTextWindow3 = 0;
   m_pMergeResultWindow = 0;
   m_pMergeWindowFrame = 0;
   m_bOutputModified = false;
   
   
   // Option handling
   KCmdLineArgs *args = KCmdLineArgs::parsedArgs();

   m_outputFilename = args->getOption("output");
   if ( m_outputFilename.isEmpty() )
      m_outputFilename = args->getOption("out");   

   if ( m_outputFilename.isEmpty() && args->isSet("merge") )
   {
      m_outputFilename = "unnamed.txt";
      m_bDefaultFilename = true;
   }
   else
      m_bDefaultFilename = false;

   g_bAutoSolve = !args->isSet("qall");  // Note that this is effective only once.
   
   m_filename1 = args->getOption("base");
   if ( m_filename1.isEmpty() )
   {
      if ( args->count() > 0 ) m_filename1 = args->arg(0);
      if ( args->count() > 1 ) m_filename2 = args->arg(1);
      if ( args->count() > 2 ) m_filename3 = args->arg(2);
   }
   else
   {
      if ( args->count() > 0 ) m_filename2 = args->arg(0);
      if ( args->count() > 1 ) m_filename3 = args->arg(1);
   }
   
   ///////////////////////////////////////////////////////////////////
   // call inits to invoke all other construction parts
   initActions();
   initStatusBar();

   // All default values must be set before calling readOptions().
   m_pOptionDialog = new OptionDialog( this );
   connect( m_pOptionDialog, SIGNAL(applyClicked()), this, SLOT(slotRefresh()) );

   readOptions();

   m_pMainSplitter = new QSplitter(this);
   m_pMainSplitter->setOrientation( Vertical );
   setCentralWidget( m_pMainSplitter );
   m_pDirectoryMergeSplitter = new QSplitter( m_pMainSplitter );
   m_pDirectoryMergeSplitter->setOrientation( Horizontal );
   m_pDirectoryMergeWindow = new DirectoryMergeWindow( m_pDirectoryMergeSplitter, m_pOptionDialog,
      KApplication::kApplication()->iconLoader() );
   m_pDirectoryMergeInfo = new DirectoryMergeInfo( m_pDirectoryMergeSplitter );
   m_pDirectoryMergeWindow->setDirectoryMergeInfo( m_pDirectoryMergeInfo );
   connect( m_pDirectoryMergeWindow, SIGNAL(startDiffMerge(QString,QString,QString,QString)),
            this, SLOT( slotFileOpen2(QString,QString,QString,QString)));
   connect( m_pDirectoryMergeWindow, SIGNAL(selectionChanged()), this, SLOT(slotUpdateAvailabilities()));
   connect( m_pDirectoryMergeWindow, SIGNAL(currentChanged(QListViewItem*)), this, SLOT(slotUpdateAvailabilities()));
   connect( m_pDirectoryMergeWindow, SIGNAL(checkIfCanContinue(bool*)), this, SLOT(slotCheckIfCanContinue(bool*)));
   connect( m_pDirectoryMergeWindow, SIGNAL(updateAvailabilities()), this, SLOT(slotUpdateAvailabilities()));

   initDirectoryMergeActions();

   m_bDirCompare = improveFilenames();
   if (!m_bDirCompare)
   {
      m_pDirectoryMergeSplitter->hide();

      init( m_filename1, m_filename2, m_filename3 );
   }

   statusBar()->setSizeGripEnabled(false);


   ///////////////////////////////////////////////////////////////////
   // disable actions at startup
   fileSave->setEnabled(false);
   fileSaveAs->setEnabled(false);

   //editCut->setEnabled(false);
   //editCopy->setEnabled(false);
   //editPaste->setEnabled(false);
   if ( !m_bDirCompare )
   {
      bool bFileOpenError = false;
      if ( ! m_filename1.isEmpty() && m_pBuf1==0  ||
           ! m_filename2.isEmpty() && m_pBuf2==0  ||
           ! m_filename3.isEmpty() && m_pBuf3==0 )
      {
         QString text( i18n("Opening of these files failed:") );
         text += "\n\n";
         if ( ! m_filename1.isEmpty() && m_pBuf1==0 )
            text += " - " + m_filename1 + "\n";
         if ( ! m_filename2.isEmpty() && m_pBuf2==0 )
            text += " - " + m_filename2 + "\n";
         if ( ! m_filename3.isEmpty() && m_pBuf3==0 )
            text += " - " + m_filename3 + "\n";

         KMessageBox::sorry( this, text, i18n("File open error") );

         bFileOpenError = true;
      }

      if ( m_filename1.isEmpty() || m_filename2.isEmpty() || bFileOpenError )
         slotFileOpen();
   }
   slotClipboardChanged(); // For initialisation.
   
   dirShowBoth->setChecked( true );
   slotUpdateAvailabilities();

   args->clear(); // Free up some memory.
}

KDiff3App::~KDiff3App()
{

}

void KDiff3App::initActions()
{
   fileOpen = KStdAction::open(this, SLOT(slotFileOpen()), actionCollection());
   fileSave = KStdAction::save(this, SLOT(slotFileSave()), actionCollection());
   fileSaveAs = KStdAction::saveAs(this, SLOT(slotFileSaveAs()), actionCollection());
   fileQuit = KStdAction::quit(this, SLOT(slotFileQuit()), actionCollection());
   editCut = KStdAction::cut(this, SLOT(slotEditCut()), actionCollection());
   editCopy = KStdAction::copy(this, SLOT(slotEditCopy()), actionCollection());
   editPaste = KStdAction::paste(this, SLOT(slotEditPaste()), actionCollection());
   viewToolBar = KStdAction::showToolbar(this, SLOT(slotViewToolBar()), actionCollection());
   viewStatusBar = KStdAction::showStatusbar(this, SLOT(slotViewStatusBar()), actionCollection());
   KStdAction::preferences(this, SLOT(slotConfigure()), actionCollection() );

   fileOpen->setStatusText(i18n("Opens documents for comparison ..."));
   fileSave->setStatusText(i18n("Saves the merge result. All conflicts must be solved!"));
   fileSaveAs->setStatusText(i18n("Saves the current document as..."));
   fileQuit->setStatusText(i18n("Quits the application"));
   editCut->setStatusText(i18n("Cuts the selected section and puts it to the clipboard"));
   editCopy->setStatusText(i18n("Copies the selected section to the clipboard"));
   editPaste->setStatusText(i18n("Pastes the clipboard contents to actual position"));
   viewToolBar->setStatusText(i18n("Enables/disables the toolbar"));
   viewStatusBar->setStatusText(i18n("Enables/disables the statusbar"));

#include "downend.xpm"
#include "currentpos.xpm"
#include "down1arrow.xpm"
#include "down2arrow.xpm"
#include "upend.xpm"
#include "up1arrow.xpm"
#include "up2arrow.xpm"
#include "iconA.xpm"
#include "iconB.xpm"
#include "iconC.xpm"
#include "autoadvance.xpm"
//#include "reload.xpm"
    
   goCurrent = new KAction(i18n("Go to Current Delta"), QIconSet(QPixmap(currentpos)), CTRL+Key_Space, this, SLOT(slotGoCurrent()), actionCollection(), "go_current");
   goTop = new KAction(i18n("Go to First Delta"), QIconSet(QPixmap(upend)), 0, this, SLOT(slotGoTop()), actionCollection(), "go_top");
   goBottom = new KAction(i18n("Go to Last Delta"), QIconSet(QPixmap(downend)), 0, this, SLOT(slotGoBottom()), actionCollection(), "go_bottom");
   goPrevDelta = new KAction(i18n("Go to PrevDelta"), QIconSet(QPixmap(up1arrow)), CTRL+Key_Up, this, SLOT(slotGoPrevDelta()), actionCollection(), "go_prev_delta");
   goNextDelta = new KAction(i18n("Go to NextDelta"), QIconSet(QPixmap(down1arrow)), CTRL+Key_Down, this, SLOT(slotGoNextDelta()), actionCollection(), "go_next_delta");
   goPrevConflict = new KAction(i18n("Go to Previous Conflict"), QIconSet(QPixmap(up2arrow)), CTRL+Key_PageUp, this, SLOT(slotGoPrevConflict()), actionCollection(), "go_prev_conflict");
   goNextConflict = new KAction(i18n("Go to Next Conflict"), QIconSet(QPixmap(down2arrow)), CTRL+Key_PageDown, this, SLOT(slotGoNextConflict()), actionCollection(), "go_next_conflict");
   chooseA = new KToggleAction(i18n("Select line(s) from A"), QIconSet(QPixmap(iconA)), CTRL+Key_1, this, SLOT(slotChooseA()), actionCollection(), "choose_a");
   chooseB = new KToggleAction(i18n("Select line(s) from B"), QIconSet(QPixmap(iconB)), CTRL+Key_2, this, SLOT(slotChooseB()), actionCollection(), "choose_b");
   chooseC = new KToggleAction(i18n("Select line(s) from C"), QIconSet(QPixmap(iconC)), CTRL+Key_3, this, SLOT(slotChooseC()), actionCollection(), "choose_c");
   autoAdvance = new KToggleAction(i18n("Automatically go to next conflict after source selection"), QIconSet(QPixmap(autoadvance)), 0, actionCollection(), "autoadvance");
   chooseAEverywhere = new KAction(i18n("Choose A Everywhere"), 0, this, SLOT(slotChooseAEverywhere()), actionCollection(), "choose_a_everywhere");
   chooseBEverywhere = new KAction(i18n("Choose B Everywhere"), 0, this, SLOT(slotChooseBEverywhere()), actionCollection(), "choose_b_everywhere");
   chooseCEverywhere = new KAction(i18n("Choose C Everywhere"), 0, this, SLOT(slotChooseCEverywhere()), actionCollection(), "choose_c_everywhere");
   autoSolve = new KAction(i18n("Automatically solve simple conflicts"), 0, this, SLOT(slotAutoSolve()), actionCollection(), "autosolve");
   unsolve = new KAction(i18n("Set deltas to conflicts"), 0, this, SLOT(slotUnsolve()), actionCollection(), "autounsolve");
   fileReload  = new KAction(i18n("Reload"), /*QIconSet(QPixmap(reloadIcon)),*/ 0, this, SLOT(slotReload()), actionCollection(), "file_reload");
}

void KDiff3App::initDirectoryMergeActions()
{
#include "startmerge.xpm"
   //dirOpen = new KAction(i18n("Open directories ..."), 0, this, SLOT(slotDirOpen()), actionCollection(), "dir_open");
   dirStartOperation = new KAction(i18n("Start/Continue directory merge"), Key_F5, m_pDirectoryMergeWindow, SLOT(mergeContinue()), actionCollection(), "dir_start_operation");
   dirCompareCurrent = new KAction(i18n("Compare selected file"), 0, m_pDirectoryMergeWindow, SLOT(compareCurrentFile()), actionCollection(), "dir_compare_current");
   dirMergeCurrent = new KAction(i18n("Merge selected file"), QIconSet(QPixmap(startmerge)), 0, m_pDirectoryMergeWindow, SLOT(mergeCurrentFile()), actionCollection(), "dir_merge_current");
   dirShowBoth = new KToggleAction(i18n("Split Screen View"), 0, this, SLOT(slotDirShowBoth()), actionCollection(), "dir_show_both");
   dirViewToggle = new KAction(i18n("Toggle View"), 0, this, SLOT(slotDirViewToggle()), actionCollection(), "dir_view_toggle");
   dirRescan = new KAction(i18n("Rescan"), 0, m_pDirectoryMergeWindow, SLOT(reload()), actionCollection(), "dir_rescan");
   dirChooseAEverywhere = new KAction(i18n("Choose A for all items"), 0, m_pDirectoryMergeWindow, SLOT(slotChooseAEverywhere()), actionCollection(), "dir_choose_a_everywhere");
   dirChooseBEverywhere = new KAction(i18n("Choose B for all items"), 0, m_pDirectoryMergeWindow, SLOT(slotChooseBEverywhere()), actionCollection(), "dir_choose_b_everywhere");
   dirChooseCEverywhere = new KAction(i18n("Choose C for all items"), 0, m_pDirectoryMergeWindow, SLOT(slotChooseCEverywhere()), actionCollection(), "dir_choose_c_everywhere");
   dirAutoChoiceEverywhere = new KAction(i18n("Auto-choose operation for all items"), 0, m_pDirectoryMergeWindow, SLOT(slotAutoChooseEverywhere()), actionCollection(), "dir_autochoose_everywhere");
   dirDoNothingEverywhere = new KAction(i18n("No operation for all items"), 0, m_pDirectoryMergeWindow, SLOT(slotNoOpEverywhere()), actionCollection(), "dir_nothing_everywhere");
   // expand all
   // collapse all
   // choose A/B/C/Suggestion/NoOp everywhere

   // use the absolute path to your kdiff3ui.rc file for testing purpose in createGUI();
   createGUI();
}

void KDiff3App::initStatusBar()
{
  ///////////////////////////////////////////////////////////////////
  // STATUSBAR
  statusBar()->message( i18n("Ready.") );
}

void KDiff3App::saveOptions()
{
   config->setGroup("General Options");
   config->writeEntry("Geometry", size());
   config->writeEntry("Position", pos());
   config->writeEntry("Show Toolbar", viewToolBar->isChecked());
   config->writeEntry("Show Statusbar",viewStatusBar->isChecked());
   config->writeEntry("ToolBarPos", (int) toolBar("mainToolBar")->barPos());

   m_pOptionDialog->m_bAutoAdvance = autoAdvance->isChecked();
   m_pOptionDialog->saveOptions( config );
}


void KDiff3App::readOptions()
{
   config->setGroup("General Options");

   // bar status settings
   bool bViewToolbar = config->readBoolEntry("Show Toolbar", true);
   viewToolBar->setChecked(bViewToolbar);
   slotViewToolBar();

   bool bViewStatusbar = config->readBoolEntry("Show Statusbar", true);
   viewStatusBar->setChecked(bViewStatusbar);
   slotViewStatusBar();


   // bar position settings
   KToolBar::BarPosition toolBarPos;
   toolBarPos=(KToolBar::BarPosition) config->readNumEntry("ToolBarPos", KToolBar::Top);
   toolBar("mainToolBar")->setBarPos(toolBarPos);

   QSize size=config->readSizeEntry("Geometry");
   QPoint pos=config->readPointEntry("Position");
   if(!size.isEmpty())
   {
      resize( size );
      move( pos );
   }

   m_pOptionDialog->readOptions( config );

   slotRefresh();
}


bool KDiff3App::queryClose()
{
   if(m_bOutputModified)
   {
      int result = KMessageBox::warningYesNoCancel(this,
         i18n("The merge result hasn't been saved."),
         i18n("Warning"), i18n("Save and quit"), i18n("Don't save and quit") );
      if ( result==KMessageBox::Cancel )
         return false;
      else if ( result==KMessageBox::Yes )
      {
         slotFileSave();
         if ( m_bOutputModified )
         {
            KMessageBox::sorry(this, i18n("Saving the merge result failed."), i18n("Warning") );
            return false;
         }
      }
   }

   m_bOutputModified = false;

   if ( m_pDirectoryMergeWindow->isDirectoryMergeInProgress() )
   {
      int result = KMessageBox::warningYesNo(this,
         i18n("You are currently doing a directory merge. Are you sure, you want to abort?"),
         i18n("Warning"), i18n("Yes - Quit"), i18n("No - Continue merging") );
      if ( result!=KMessageBox::Yes )
         return false;
   }

   return true;
}

bool KDiff3App::queryExit()
{
  saveOptions();
  return true;
}

/////////////////////////////////////////////////////////////////////
// SLOT IMPLEMENTATION
/////////////////////////////////////////////////////////////////////


void KDiff3App::slotFileSave()
{
   if ( m_bDefaultFilename )
   {
      slotFileSaveAs();
   }
   else
   {
      slotStatusMsg(i18n("Saving file..."));

      bool bSuccess = m_pMergeResultWindow->saveDocument( m_outputFilename );
      if ( bSuccess )
      {
         m_bOutputModified = false;
         if ( m_bDirCompare )
            m_pDirectoryMergeWindow->mergeResultSaved(m_outputFilename);
      }

      slotStatusMsg(i18n("Ready."));
   }
}

void KDiff3App::slotFileSaveAs()
{
  slotStatusMsg(i18n("Saving file with a new filename..."));

  QString s = KFileDialog::getSaveFileName( QDir::currentDirPath(),
        0, this, i18n("Save as...") );
  if(!s.isEmpty())
  {
     m_outputFilename = s;
     bool bSuccess = m_pMergeResultWindow->saveDocument( m_outputFilename );
     if ( bSuccess )
     {
        m_bOutputModified = false;
        if ( m_bDirCompare )
           m_pDirectoryMergeWindow->mergeResultSaved(m_outputFilename);
     }
     //setCaption(url.fileName(),doc->isModified());

     m_bDefaultFilename = false;
  }

  slotStatusMsg(i18n("Ready."));
}

void KDiff3App::slotFileClose()
{
  slotStatusMsg(i18n("Closing file..."));

  close();

  slotStatusMsg(i18n("Ready."));
}


void KDiff3App::slotFileQuit()
{
   slotStatusMsg(i18n("Exiting..."));
   saveOptions();

   // close the first window, the list makes the next one the first again.
   // This ensures that queryClose() is called on each window to ask for closing
   KMainWindow* w;
   if(memberList)
   {
     for(w=memberList->first(); w!=0; w=memberList->next())
     {
       // only close the window if the closeEvent is accepted. If the user presses Cancel on the saveModified() dialog,
       // the window and the application stay open.
       if(!w->close())
          return;      // Don't quit
     }
   }
   KApplication::exit(0);
}



void KDiff3App::slotViewToolBar()
{
  slotStatusMsg(i18n("Toggling toolbar..."));
  ///////////////////////////////////////////////////////////////////
  // turn Toolbar on or off
  if(!viewToolBar->isChecked())
  {
    toolBar("mainToolBar")->hide();
  }
  else
  {
    toolBar("mainToolBar")->show();
  }

  slotStatusMsg(i18n("Ready."));
}

void KDiff3App::slotViewStatusBar()
{
  slotStatusMsg(i18n("Toggle the statusbar..."));
  ///////////////////////////////////////////////////////////////////
  //turn Statusbar on or off
  if(!viewStatusBar->isChecked())
  {
    statusBar()->hide();
  }
  else
  {
    statusBar()->show();
  }

  slotStatusMsg(i18n("Ready."));
}


void KDiff3App::slotStatusMsg(const QString &text)
{
  ///////////////////////////////////////////////////////////////////
  // change status message permanently
  statusBar()->clear();
  statusBar()->message( text );
}
